# ------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License (MIT). See License.txt in the repo root for license information.
# ------------------------------------------------------------

[CmdletBinding(DefaultParametersetName="Unsecure")] 
param (
    [Parameter(ParameterSetName="Unsecure", Mandatory=$true)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$true)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$true)]
    [string] $NodeName,

    [Parameter(ParameterSetName="Unsecure", Mandatory=$true)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$true)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$true)]
    [string] $NodeType,

    [Parameter(ParameterSetName="Unsecure", Mandatory=$true)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$true)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$true)]
    [string] $NodeIpAddressOrFQDN,

    [Parameter(ParameterSetName="Unsecure", Mandatory=$true)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$true)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$true)]
    [string] $ExistingClientConnectionEndpoint,

    [Parameter(ParameterSetName="Unsecure", Mandatory=$true)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$true)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$true)]
    [string] $UpgradeDomain,

    [Parameter(ParameterSetName="Unsecure", Mandatory=$true)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$true)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$true)]
    [string] $FaultDomain,

    [Parameter(ParameterSetName="Unsecure", Mandatory=$false)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$false)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$false)]
    [switch] $AcceptEULA,

    [Parameter(ParameterSetName="Unsecure", Mandatory=$false)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$false)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$false)]
    [switch] $Force,

    [Parameter(ParameterSetName="Unsecure", Mandatory=$false)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$false)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$false)]
    [switch] $NoCleanupOnFailure,

    [Parameter(ParameterSetName="Unsecure", Mandatory=$false)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$false)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$false)]
    [switch] $BypassUpgradeStateValidation,

    [Parameter(ParameterSetName="Unsecure", Mandatory=$false)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$false)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$false)]
    [string] $FabricRuntimePackagePath,

    [Parameter(ParameterSetName="Unsecure", Mandatory=$false)]
    [Parameter(ParameterSetName="Certificate", Mandatory=$false)]
    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$false)]
    [int] $TimeoutInSeconds,

    [Parameter(ParameterSetName="Certificate", Mandatory=$true)]
    [switch] $X509Credential,

    [Parameter(ParameterSetName="Certificate", Mandatory=$true)]
    [string] $ServerCertThumbprint,

    [Parameter(ParameterSetName="Certificate", Mandatory=$true)]
    [string] $StoreLocation,

    [Parameter(ParameterSetName="Certificate", Mandatory=$true)]
    [string] $StoreName,

    [Parameter(ParameterSetName="Certificate", Mandatory=$true)]
    [string] $FindValueThumbprint,

    [Parameter(ParameterSetName="WindowsSecurity", Mandatory=$true)]
    [switch] $WindowsCredential

)

$Identity = [System.Security.Principal.WindowsIdentity]::GetCurrent()
$Principal = New-Object System.Security.Principal.WindowsPrincipal($Identity)
$IsAdmin = $Principal.IsInRole([System.Security.Principal.WindowsBuiltInRole]::Administrator)

if(!$IsAdmin)
{
    Write-Host "Please run the script with administrative privileges." -ForegroundColor "Red"
    exit 1
}

if(!$AcceptEULA.IsPresent)
{
    $EulaAccepted = Read-Host 'Do you accept the license terms for using Microsoft Azure Service Fabric located in the root of your package download? If you do not accept the license terms you may not use the software.
[Y] Yes  [N] No  [?] Help (default is "N")'
    if($EulaAccepted -ne "y" -and $EulaAccepted -ne "Y")
    {
        Write-host "You need to accept the license terms for using Microsoft Azure Service Fabric located in the root of your package download before you can use the software." -ForegroundColor "Red"
        exit 1
    }
}

$ThisScriptPath = $(Split-Path -parent $MyInvocation.MyCommand.Definition)
$DeployerBinPath = Join-Path $ThisScriptPath -ChildPath "DeploymentComponents"
if(!(Test-Path $DeployerBinPath))
{
    $DCAutoExtractorPath = Join-Path $ThisScriptPath "DeploymentComponentsAutoextractor.exe"
    if(!(Test-Path $DCAutoExtractorPath)) 
    {
        Write-Host "Standalone package DeploymentComponents and DeploymentComponentsAutoextractor.exe are not present local to the script location."
        exit 1
    }

    #Extract DeploymentComponents
    $DCExtractArguments = "/E /Y /L `"$ThisScriptPath`""
    $DCExtractOutput = cmd.exe /c "$DCAutoExtractorPath $DCExtractArguments && exit 0 || exit 1"
    if($LASTEXITCODE -eq 1)
    {
        Write-Host "Extracting DeploymentComponents Cab ran into an issue."
        Write-Host $DCExtractOutput
        exit 1
    }
    else
    {
        Write-Host "DeploymentComponents extracted."
    }
}

$SystemFabricModulePath = Join-Path $DeployerBinPath -ChildPath "System.Fabric.dll"
if(!(Test-Path $SystemFabricModulePath)) 
{
    Write-Host "Run the script local to the Standalone package directory."
    exit 1
}

$MicrosoftServiceFabricCabFileAbsolutePath = $null
if ((Test-Path variable:FabricRuntimePackagePath) `
        -and ![string]::IsNullOrEmpty($FabricRuntimePackagePath))
{
    $MicrosoftServiceFabricCabFileAbsolutePath = Resolve-Path $FabricRuntimePackagePath
    if(!(Test-Path $MicrosoftServiceFabricCabFileAbsolutePath)) 
    {
        Write-Host "Microsoft Service Fabric Runtime package not found in the specified directory : $FabricRuntimePackagePath"
        exit 1
    }

    Write-Verbose "Using runtime package $MicrosoftServiceFabricCabFileAbsolutePath"
}
else
{
    $RuntimeBinPath = Join-Path $ThisScriptPath -ChildPath "DeploymentRuntimePackages"
    if(!(Test-Path $RuntimeBinPath)) 
    {
        Write-Host "No directory exists for Runtime packages. Creating a new directory."
        md $RuntimeBinPath | Out-Null
        Write-Host "Done creating $RuntimeBinPath"
    }

    Write-Verbose "Will use downloaded runtime package for deployment from $RuntimeBinPath"
}

$ServiceFabricPowershellModulePath = Join-Path $DeployerBinPath -ChildPath "ServiceFabric.psd1"

if (!(Test-Path variable:TimeoutInSeconds) `
    -or ($TimeoutInSeconds -le 0))
{
    $TimeoutInSeconds = 300
    Write-Verbose "TimeoutInSeconds was not set. Defaulting to $TimeoutInSeconds seconds."
}

$parentVerbosePreference = $VerbosePreference

Write-Verbose "MicrosoftServiceFabricCabFileAbsolutePath $MicrosoftServiceFabricCabFileAbsolutePath"

# Invoke in separate AppDomain
if($X509Credential)
{
    Write-Verbose "X509Credential: $X509Credential"
    $argList = @( $DeployerBinPath, $ExistingClientConnectionEndpoint, $ServiceFabricPowershellModulePath, $NodeName, $NodeType, $NodeIpAddressOrFQDN, $UpgradeDomain, $FaultDomain, $Force.IsPresent, $NoCleanupOnFailure.IsPresent, $BypassUpgradeStateValidation.IsPresent, $MicrosoftServiceFabricCabFileAbsolutePath, $TimeoutInSeconds, $parentVerbosePreference, $false, $true, $ServerCertThumbprint, $StoreLocation, $StoreName, $FindValueThumbprint )
}
else
{
    if($WindowsCredential)
    {
        Write-Verbose "WindowsCredential: $WindowsCredential"
        $argList = @( $DeployerBinPath, $ExistingClientConnectionEndpoint, $ServiceFabricPowershellModulePath, $NodeName, $NodeType, $NodeIpAddressOrFQDN, $UpgradeDomain, $FaultDomain, $Force.IsPresent, $NoCleanupOnFailure.IsPresent, $BypassUpgradeStateValidation.IsPresent, $MicrosoftServiceFabricCabFileAbsolutePath, $TimeoutInSeconds, $parentVerbosePreference, $true )
    }
    else
    {
        Write-Verbose "Not X509Credential nor WindowsCredential."
        $argList = @( $DeployerBinPath, $ExistingClientConnectionEndpoint, $ServiceFabricPowershellModulePath, $NodeName, $NodeType, $NodeIpAddressOrFQDN, $UpgradeDomain, $FaultDomain, $Force.IsPresent, $NoCleanupOnFailure.IsPresent, $BypassUpgradeStateValidation.IsPresent, $MicrosoftServiceFabricCabFileAbsolutePath, $TimeoutInSeconds, $parentVerbosePreference )
    }
}

Powershell -Command {
    param (
        [Parameter(Mandatory=$true)]
        [string] $DeployerBinPath,
        
        [Parameter(Mandatory=$true)]
        [string] $ExistingClientConnectionEndpoint,

        [Parameter(Mandatory=$true)]
        [string] $ServiceFabricPowershellModulePath,

        [Parameter(Mandatory=$true)]
        [string] $NodeName,

        [Parameter(Mandatory=$true)]
        [string] $NodeType,

        [Parameter(Mandatory=$true)]
        [string] $NodeIpAddressOrFQDN,

        [Parameter(Mandatory=$true)]
        [string] $UpgradeDomain,

        [Parameter(Mandatory=$true)]
        [string] $FaultDomain,

        [Parameter(Mandatory=$false)]
        [bool] $Force,

        [Parameter(Mandatory=$false)]
        [bool] $NoCleanupOnFailure,

        [Parameter(Mandatory=$false)]
        [bool] $BypassUpgradeStateValidation,

        [Parameter(Mandatory=$false)]
        [string] $MicrosoftServiceFabricCabFileAbsolutePath,

        [Parameter(Mandatory=$false)]
        [int] $TimeoutInSeconds,

        [Parameter(Mandatory=$false)]
        [string] $ParentVerbosePreference,

        [Parameter(Mandatory=$false)]
        [bool] $WindowsCredential,
        
        [Parameter(Mandatory=$false)]
        [bool] $X509Credential,

        [Parameter(Mandatory=$false)]
        [string] $ServerCertThumbprint,

        [Parameter(Mandatory=$false)]
        [string] $StoreLocation,

        [Parameter(Mandatory=$false)]
        [string] $StoreName,

        [Parameter(Mandatory=$false)]
        [string] $FindValueThumbprint
    )
    
    #Add FabricCodePath Environment Path
    $env:path = "$($DeployerBinPath);" + $env:path

    #Import Service Fabric Powershell Module
    Import-Module $ServiceFabricPowershellModulePath

    Try
    {
        # Connect to the existing cluster
        Write-Verbose "Connecting to the cluster at $ExistingClientConnectionEndpoint"
        if($X509Credential)
        {
            Connect-ServiceFabricCluster -ConnectionEndpoint $ExistingClientConnectionEndpoint -X509Credential -ServerCertThumbprint $ServerCertThumbprint -StoreLocation $StoreLocation -StoreName $StoreName -FindValue $FindValueThumbprint -FindType FindByThumbprint
        }
        else
        {
            if($WindowsCredential)
            {
                Connect-ServiceFabricCluster $ExistingClientConnectionEndpoint -WindowsCredential
            }
            else
            {
                Connect-ServiceFabricCluster $ExistingClientConnectionEndpoint
            }
        }
    }
    Catch
    {
        Write-Host "Cannot form client connection to cluster. Check your ClientConnectionEndpoint and security credentials. Exception thrown : $($_.Exception.ToString())" -ForegroundColor Red
        exit 1
    }

    if ($ClusterConnection -eq $null) {
        Write-Host "Could not form client connection to cluster. Check your ClientConnectionEndpoint and security credentials." -ForegroundColor Red
        exit 1
    }

    if (!(Test-Path variable:MicrosoftServiceFabricCabFileAbsolutePath) `
            -or [string]::IsNullOrEmpty($MicrosoftServiceFabricCabFileAbsolutePath))
    {
        Try
        {
            # Get runtime package details
            $UpgradeStatus = Get-ServiceFabricClusterUpgrade
        }
        Catch
        {
            Write-Host "Could not query current cluster version. Check your ClientConnectionEndpoint and security credentials." -ForegroundColor Red
            exit 1
        }

        if ($UpgradeStatus.UpgradeState -ne "RollingForwardCompleted" -And $UpgradeStatus.UpgradeState -ne "RollingBackCompleted")
        {
            if ($BypassUpgradeStateValidation)
            {
                Write-Host "BypassUpgradeStateValidation is set but cannot be used without FabricRuntimePackagePath when cluster version cannot be automatically inferred from the upgrade state." -ForegroundColor Red
            }

            Write-Host "New node cannot be added to the cluster while upgrade is in progress or before cluster has finished bootstrapping. To monitor upgrade state run Get-ServiceFabricClusterUpgrade and wait until UpgradeState switches to either RollingForwardCompleted or RollingBackCompleted." -ForegroundColor Red
            exit 1
        }

        Try
        {
            $RuntimeCabFilename = "MicrosoftAzureServiceFabric." + $UpgradeStatus.TargetCodeVersion + ".cab"
            $DeploymentPackageRoot = Split-Path -parent $DeployerBinPath
            $RuntimeBinPath = Join-Path $DeploymentPackageRoot -ChildPath "DeploymentRuntimePackages"
            $MicrosoftServiceFabricCabFilePath = Join-Path $RuntimeBinPath -ChildPath $RuntimeCabFilename
                
            if (!(Test-Path $MicrosoftServiceFabricCabFilePath))
            {
                $RuntimePackageDetails = Get-ServiceFabricRuntimeSupportedVersion
                $RequiredPackage = $RuntimePackageDetails.RuntimePackages | where { $_.Version -eq $UpgradeStatus.TargetCodeVersion }
                    
                if ($RequiredPackage -eq $null)
                {
                    Write-Host "The required runtime version is no longer supported. Please upgrade your cluster to the latest version before adding a node." -ForegroundColor Red
                    exit 1
                }
                $Version = $UpgradeStatus.TargetCodeVersion
                Write-Host "Runtime package version $Version was not found in DeploymentRuntimePackages folder and needed to be downloaded."
                (New-Object System.Net.WebClient).DownloadFile($RuntimePackageDetails.GoalRuntimeLocation, $MicrosoftServiceFabricCabFilePath)
                Write-Host "Runtime package has been successfully downloaded to $MicrosoftServiceFabricCabFilePath."
            }
            $MicrosoftServiceFabricCabFileAbsolutePath = Resolve-Path $MicrosoftServiceFabricCabFilePath
        }
        Catch
        {
            Write-Host "Runtime package cannot be downloaded. Check your internet connectivity. If the cluster is not connected to the internet run Get-ServiceFabricClusterUpgrade and note the TargetCodeVersion. Run Get-ServiceFabricRuntimeSupportedVersion from a machine connected to the internet to get the download links for all supported fabric versions. Download the package corresponding to your TargetCodeVersion. Pass -FabricRuntimePackageOutputDirectory <Path to runtime package> to AddNode.ps1 in addition to other parameters. Exception thrown : $($_.Exception.ToString())" -ForegroundColor Red
            exit 1
        }
    }

    #Add Node to an existing cluster
    Try
    {
        $AddNodeTimeoutDuration=[System.TimeSpan]::FromSeconds($TimeoutInSeconds)
        $AddNodeTimeout=[System.DateTime]::UtcNow + $AddNodeTimeoutDuration

        Get-ServiceFabricNode | ft

        $VerbosePreference = $ParentVerbosePreference
        $params = @{
                        'NodeName' = $NodeName;
                        'NodeType' = $NodeType;
                        'IpAddressOrFQDN' = $NodeIpAddressOrFQDN;
                        'UpgradeDomain' = $UpgradeDomain;
                        'FaultDomain' = $FaultDomain;
                        'FabricRuntimePackagePath' = $MicrosoftServiceFabricCabFileAbsolutePath;
                        'NoCleanupOnFailure' = $NoCleanupOnFailure;
                        'Force' = $Force;
                        'BypassUpgradeStateValidation' = $BypassUpgradeStateValidation;
                    }

        if ((Test-Path variable:TimeoutInSeconds) `
            -and ($TimeoutInSeconds -gt 0))
        {
            $params += @{'TimeoutSec' = $TimeoutInSeconds;}
        }


        Write-Host "Adding Node $NodeName"
        Add-ServiceFabricNode @params

        Write-Host "Waiting for node to join cluster." -NoNewline

        $nodeJoined=$false
        $firstRun=$true
        do
        {
            if (!$nodeJoined -and !$firstRun)
            {
                Write-Host "." -NoNewline
                Start-Sleep -s 10
            }

            $result=Get-ServiceFabricNode | Where-Object { [System.String]::Compare($_.NodeName, $NodeName, [System.StringComparison]::InvariantCulture) -eq 0 }
            $nodeJoined=$result.Count -gt 0
            $firstRun=$false
        } while(!$nodeJoined -and ([System.DateTime]::UtcNow -lt $AddNodeTimeout))

        Write-Host "" # Newline
        if(-not $nodeJoined)
        {
            Write-Host "Node did not join within timeout of $TimeoutInSeconds seconds." -ForegroundColor Red
            exit 1
        }

        Get-ServiceFabricNode | ft

        Write-Host "Node $NodeName joined the cluster!" -ForegroundColor Green
    }
    Catch
    {
        if($VerbosePreference -eq "SilentlyContinue")
        {
            Write-Host "Add Node failed. Call with -Verbose for more details" -ForegroundColor Red
        }
        Write-Host "Add node to existing cluster failed with exception: $($_.Exception.ToString())" -ForegroundColor Red
        exit 1
    }
    
} -args $argList -OutputFormat Text

$env:Path = [System.Environment]::GetEnvironmentVariable("path","Machine")
# SIG # Begin signature block
# MIIjlAYJKoZIhvcNAQcCoIIjhTCCI4ECAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCAZK+1P+seFOB8R
# U92DWOXsr2WUJM+ZzPvyllpHNGh1VKCCDXYwggX0MIID3KADAgECAhMzAAABhk0h
# daDZB74sAAAAAAGGMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjAwMzA0MTgzOTQ2WhcNMjEwMzAzMTgzOTQ2WjB0MQsw
# CQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9u
# ZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMR4wHAYDVQQDExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIB
# AQC49eyyaaieg3Xb7ew+/hA34gqzRuReb9svBF6N3+iLD5A0iMddtunnmbFVQ+lN
# Wphf/xOGef5vXMMMk744txo/kT6CKq0GzV+IhAqDytjH3UgGhLBNZ/UWuQPgrnhw
# afQ3ZclsXo1lto4pyps4+X3RyQfnxCwqtjRxjCQ+AwIzk0vSVFnId6AwbB73w2lJ
# +MC+E6nVmyvikp7DT2swTF05JkfMUtzDosktz/pvvMWY1IUOZ71XqWUXcwfzWDJ+
# 96WxBH6LpDQ1fCQ3POA3jCBu3mMiB1kSsMihH+eq1EzD0Es7iIT1MlKERPQmC+xl
# K+9pPAw6j+rP2guYfKrMFr39AgMBAAGjggFzMIIBbzAfBgNVHSUEGDAWBgorBgEE
# AYI3TAgBBggrBgEFBQcDAzAdBgNVHQ4EFgQUhTFTFHuCaUCdTgZXja/OAQ9xOm4w
# RQYDVR0RBD4wPKQ6MDgxHjAcBgNVBAsTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEW
# MBQGA1UEBRMNMjMwMDEyKzQ1ODM4NDAfBgNVHSMEGDAWgBRIbmTlUAXTgqoXNzci
# tW2oynUClTBUBgNVHR8ETTBLMEmgR6BFhkNodHRwOi8vd3d3Lm1pY3Jvc29mdC5j
# b20vcGtpb3BzL2NybC9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3JsMGEG
# CCsGAQUFBwEBBFUwUzBRBggrBgEFBQcwAoZFaHR0cDovL3d3dy5taWNyb3NvZnQu
# Y29tL3BraW9wcy9jZXJ0cy9NaWNDb2RTaWdQQ0EyMDExXzIwMTEtMDctMDguY3J0
# MAwGA1UdEwEB/wQCMAAwDQYJKoZIhvcNAQELBQADggIBAEDkLXWKDtJ8rLh3d7XP
# 1xU1s6Gt0jDqeHoIpTvnsREt9MsKriVGKdVVGSJow1Lz9+9bINmPZo7ZdMhNhWGQ
# QnEF7z/3czh0MLO0z48cxCrjLch0P2sxvtcaT57LBmEy+tbhlUB6iz72KWavxuhP
# 5zxKEChtLp8gHkp5/1YTPlvRYFrZr/iup2jzc/Oo5N4/q+yhOsRT3KJu62ekQUUP
# sPU2bWsaF/hUPW/L2O1Fecf+6OOJLT2bHaAzr+EBAn0KAUiwdM+AUvasG9kHLX+I
# XXlEZvfsXGzzxFlWzNbpM99umWWMQPTGZPpSCTDDs/1Ci0Br2/oXcgayYLaZCWsj
# 1m/a0V8OHZGbppP1RrBeLQKfATjtAl0xrhMr4kgfvJ6ntChg9dxy4DiGWnsj//Qy
# wUs1UxVchRR7eFaP3M8/BV0eeMotXwTNIwzSd3uAzAI+NSrN5pVlQeC0XXTueeDu
# xDch3S5UUdDOvdlOdlRAa+85Si6HmEUgx3j0YYSC1RWBdEhwsAdH6nXtXEshAAxf
# 8PWh2wCsczMe/F4vTg4cmDsBTZwwrHqL5krX++s61sLWA67Yn4Db6rXV9Imcf5UM
# Cq09wJj5H93KH9qc1yCiJzDCtbtgyHYXAkSHQNpoj7tDX6ko9gE8vXqZIGj82mwD
# TAY9ofRH0RSMLJqpgLrBPCKNMIIHejCCBWKgAwIBAgIKYQ6Q0gAAAAAAAzANBgkq
# hkiG9w0BAQsFADCBiDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24x
# EDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlv
# bjEyMDAGA1UEAxMpTWljcm9zb2Z0IFJvb3QgQ2VydGlmaWNhdGUgQXV0aG9yaXR5
# IDIwMTEwHhcNMTEwNzA4MjA1OTA5WhcNMjYwNzA4MjEwOTA5WjB+MQswCQYDVQQG
# EwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwG
# A1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSgwJgYDVQQDEx9NaWNyb3NvZnQg
# Q29kZSBTaWduaW5nIFBDQSAyMDExMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIIC
# CgKCAgEAq/D6chAcLq3YbqqCEE00uvK2WCGfQhsqa+laUKq4BjgaBEm6f8MMHt03
# a8YS2AvwOMKZBrDIOdUBFDFC04kNeWSHfpRgJGyvnkmc6Whe0t+bU7IKLMOv2akr
# rnoJr9eWWcpgGgXpZnboMlImEi/nqwhQz7NEt13YxC4Ddato88tt8zpcoRb0Rrrg
# OGSsbmQ1eKagYw8t00CT+OPeBw3VXHmlSSnnDb6gE3e+lD3v++MrWhAfTVYoonpy
# 4BI6t0le2O3tQ5GD2Xuye4Yb2T6xjF3oiU+EGvKhL1nkkDstrjNYxbc+/jLTswM9
# sbKvkjh+0p2ALPVOVpEhNSXDOW5kf1O6nA+tGSOEy/S6A4aN91/w0FK/jJSHvMAh
# dCVfGCi2zCcoOCWYOUo2z3yxkq4cI6epZuxhH2rhKEmdX4jiJV3TIUs+UsS1Vz8k
# A/DRelsv1SPjcF0PUUZ3s/gA4bysAoJf28AVs70b1FVL5zmhD+kjSbwYuER8ReTB
# w3J64HLnJN+/RpnF78IcV9uDjexNSTCnq47f7Fufr/zdsGbiwZeBe+3W7UvnSSmn
# Eyimp31ngOaKYnhfsi+E11ecXL93KCjx7W3DKI8sj0A3T8HhhUSJxAlMxdSlQy90
# lfdu+HggWCwTXWCVmj5PM4TasIgX3p5O9JawvEagbJjS4NaIjAsCAwEAAaOCAe0w
# ggHpMBAGCSsGAQQBgjcVAQQDAgEAMB0GA1UdDgQWBBRIbmTlUAXTgqoXNzcitW2o
# ynUClTAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYwDwYD
# VR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBRyLToCMZBDuRQFTuHqp8cx0SOJNDBa
# BgNVHR8EUzBRME+gTaBLhklodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2Ny
# bC9wcm9kdWN0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3JsMF4GCCsG
# AQUFBwEBBFIwUDBOBggrBgEFBQcwAoZCaHR0cDovL3d3dy5taWNyb3NvZnQuY29t
# L3BraS9jZXJ0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFfMDNfMjIuY3J0MIGfBgNV
# HSAEgZcwgZQwgZEGCSsGAQQBgjcuAzCBgzA/BggrBgEFBQcCARYzaHR0cDovL3d3
# dy5taWNyb3NvZnQuY29tL3BraW9wcy9kb2NzL3ByaW1hcnljcHMuaHRtMEAGCCsG
# AQUFBwICMDQeMiAdAEwAZQBnAGEAbABfAHAAbwBsAGkAYwB5AF8AcwB0AGEAdABl
# AG0AZQBuAHQALiAdMA0GCSqGSIb3DQEBCwUAA4ICAQBn8oalmOBUeRou09h0ZyKb
# C5YR4WOSmUKWfdJ5DJDBZV8uLD74w3LRbYP+vj/oCso7v0epo/Np22O/IjWll11l
# hJB9i0ZQVdgMknzSGksc8zxCi1LQsP1r4z4HLimb5j0bpdS1HXeUOeLpZMlEPXh6
# I/MTfaaQdION9MsmAkYqwooQu6SpBQyb7Wj6aC6VoCo/KmtYSWMfCWluWpiW5IP0
# wI/zRive/DvQvTXvbiWu5a8n7dDd8w6vmSiXmE0OPQvyCInWH8MyGOLwxS3OW560
# STkKxgrCxq2u5bLZ2xWIUUVYODJxJxp/sfQn+N4sOiBpmLJZiWhub6e3dMNABQam
# ASooPoI/E01mC8CzTfXhj38cbxV9Rad25UAqZaPDXVJihsMdYzaXht/a8/jyFqGa
# J+HNpZfQ7l1jQeNbB5yHPgZ3BtEGsXUfFL5hYbXw3MYbBL7fQccOKO7eZS/sl/ah
# XJbYANahRr1Z85elCUtIEJmAH9AAKcWxm6U/RXceNcbSoqKfenoi+kiVH6v7RyOA
# 9Z74v2u3S5fi63V4GuzqN5l5GEv/1rMjaHXmr/r8i+sLgOppO6/8MO0ETI7f33Vt
# Y5E90Z1WTk+/gFcioXgRMiF670EKsT/7qMykXcGhiJtXcVZOSEXAQsmbdlsKgEhr
# /Xmfwb1tbWrJUnMTDXpQzTGCFXQwghVwAgEBMIGVMH4xCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNp
# Z25pbmcgUENBIDIwMTECEzMAAAGGTSF1oNkHviwAAAAAAYYwDQYJYIZIAWUDBAIB
# BQCgga4wGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQwHAYKKwYBBAGCNwIBCzEO
# MAwGCisGAQQBgjcCARUwLwYJKoZIhvcNAQkEMSIEIFbosjyKtQgCFQUo70FdkcpR
# RWpTr32h78Y1JGuK10oBMEIGCisGAQQBgjcCAQwxNDAyoBSAEgBNAGkAYwByAG8A
# cwBvAGYAdKEagBhodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20wDQYJKoZIhvcNAQEB
# BQAEggEAWX9cC+yWNqXNCpTzCLDY2eUJ2KII1Utk2xak8h6GsMvNyWBiPKIbHRu+
# BU6ihUmefZ8VToPuhAy4h6mIH1BtJ3aKqzlyw+3Ii9gf+/6MlXmU/krDyfw2Ts30
# yRqJqJekLykQqBdooUOfXHceIJm2Izb9sAoWoiHpfHE9mm6+CO/3KiGLDuZLX25n
# g54JXOq/0jvohqxxy2V+KuH4bZnbeeSkVz00lElx6Q7MyvgWElRLlUvJzeo6vylN
# AQC5n7VlGUAMMumRnXVmpJUzhnz52mQlSruUpn+b4FRJaJiaqojCxdCp1ItZTdEa
# oZzxr6VoiAz01MpMdpsY9+QP6E9tN6GCEv4wghL6BgorBgEEAYI3AwMBMYIS6jCC
# EuYGCSqGSIb3DQEHAqCCEtcwghLTAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFZBgsq
# hkiG9w0BCRABBKCCAUgEggFEMIIBQAIBAQYKKwYBBAGEWQoDATAxMA0GCWCGSAFl
# AwQCAQUABCA2c5a0c/fT1sodDUkTJhz1f1btnaE+ldm1Ilx62mLVZgIGX9zcJyk/
# GBMyMDIxMDEyMjE1MzQyNS4yMDRaMASAAgH0oIHYpIHVMIHSMQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMS0wKwYDVQQLEyRNaWNyb3NvZnQgSXJl
# bGFuZCBPcGVyYXRpb25zIExpbWl0ZWQxJjAkBgNVBAsTHVRoYWxlcyBUU1MgRVNO
# OjNCRDQtNEI4MC02OUMzMSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBT
# ZXJ2aWNloIIOTTCCBPkwggPhoAMCAQICEzMAAAE7EhuSI1ICWqUAAAAAATswDQYJ
# KoZIhvcNAQELBQAwfDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24x
# EDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlv
# bjEmMCQGA1UEAxMdTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwHhcNMjAx
# MDE1MTcyODIyWhcNMjIwMTEyMTcyODIyWjCB0jELMAkGA1UEBhMCVVMxEzARBgNV
# BAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jv
# c29mdCBDb3Jwb3JhdGlvbjEtMCsGA1UECxMkTWljcm9zb2Z0IElyZWxhbmQgT3Bl
# cmF0aW9ucyBMaW1pdGVkMSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVTTjozQkQ0LTRC
# ODAtNjlDMzElMCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZTCC
# ASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAOM21pXQ1uUwGAcBUvPC91UA
# rbFP2chd56rQsPlUSuHhsZjV3UCFIW33AqybJe3bpHq8PLkopZafgB+YQPYDkXNy
# ACQvjex5WVl1/+aAJDba2zGmTIWFg4TDweE5SsPCncDqc2WReMZw4Gllmbm5fkJ1
# gdqrdrny2UuuKgvt7BowNvBDotZ+1M5WpB/rxepaAi+WOiH6t7+Qk6uwnuPWQdmQ
# lsGKZB+O3GJ3r0TMptxALPTdPkLwml4jqLCyMW4Zar4ygCub/SRQkoxE3aVGikoE
# 0lZWw1Pke/cEs73bFoD3005NDW5om69wMSVFO/oBX3biNAKN+psiZCUINdAhRDEC
# AwEAAaOCARswggEXMB0GA1UdDgQWBBTHz4/N4uhXoIpWZAlBRbq+7+aIkTAfBgNV
# HSMEGDAWgBTVYzpcijGQ80N7fEYbxTNoWoVtVTBWBgNVHR8ETzBNMEugSaBHhkVo
# dHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNUaW1T
# dGFQQ0FfMjAxMC0wNy0wMS5jcmwwWgYIKwYBBQUHAQEETjBMMEoGCCsGAQUFBzAC
# hj5odHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vcGtpL2NlcnRzL01pY1RpbVN0YVBD
# QV8yMDEwLTA3LTAxLmNydDAMBgNVHRMBAf8EAjAAMBMGA1UdJQQMMAoGCCsGAQUF
# BwMIMA0GCSqGSIb3DQEBCwUAA4IBAQBBsPG05rk1r1UH+zEtTK/c7pA1RDH1pwFv
# 3mPzlf3lbW1Nh12AMKK/Ing2fr5lps0uPbN/VWBvkAQL76NbFWiZzWtmZBjUKPbJ
# +wifpX2qe2/M0I3BFdmNdGWT3dcthT513SO2km1XbxiK5PAXsOYlmq+mEkGceSuM
# qCQ9apcefzxe5vyhnpHKbTH4+Yf/uB3HhRPTKp1YkZySI1Yp5ic5uqlzVPlUEarm
# skYYueAPjmdKGKPvYbJGLHmRbIZQmOh9VQxSAxOoASN/TA+xRs51LIKknLiunlAh
# xLn1/BWhC4+CW7CbKeqtHKV4pEvFyp+cm9Lp5lBUF1XuPa7TyyIXMIIGcTCCBFmg
# AwIBAgIKYQmBKgAAAAAAAjANBgkqhkiG9w0BAQsFADCBiDELMAkGA1UEBhMCVVMx
# EzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoT
# FU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWljcm9zb2Z0IFJvb3Qg
# Q2VydGlmaWNhdGUgQXV0aG9yaXR5IDIwMTAwHhcNMTAwNzAxMjEzNjU1WhcNMjUw
# NzAxMjE0NjU1WjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQ
# MA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9u
# MSYwJAYDVQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMDCCASIwDQYJ
# KoZIhvcNAQEBBQADggEPADCCAQoCggEBAKkdDbx3EYo6IOz8E5f1+n9plGt0VBDV
# pQoAgoX77XxoSyxfxcPlYcJ2tz5mK1vwFVMnBDEfQRsalR3OCROOfGEwWbEwRA/x
# YIiEVEMM1024OAizQt2TrNZzMFcmgqNFDdDq9UeBzb8kYDJYYEbyWEeGMoQedGFn
# kV+BVLHPk0ySwcSmXdFhE24oxhr5hoC732H8RsEnHSRnEnIaIYqvS2SJUGKxXf13
# Hz3wV3WsvYpCTUBR0Q+cBj5nf/VmwAOWRH7v0Ev9buWayrGo8noqCjHw2k4GkbaI
# CDXoeByw6ZnNPOcvRLqn9NxkvaQBwSAJk3jN/LzAyURdXhacAQVPIk0CAwEAAaOC
# AeYwggHiMBAGCSsGAQQBgjcVAQQDAgEAMB0GA1UdDgQWBBTVYzpcijGQ80N7fEYb
# xTNoWoVtVTAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMAQTALBgNVHQ8EBAMCAYYw
# DwYDVR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBTV9lbLj+iiXGJo0T2UkFvXzpoY
# xDBWBgNVHR8ETzBNMEugSaBHhkVodHRwOi8vY3JsLm1pY3Jvc29mdC5jb20vcGtp
# L2NybC9wcm9kdWN0cy9NaWNSb29DZXJBdXRfMjAxMC0wNi0yMy5jcmwwWgYIKwYB
# BQUHAQEETjBMMEoGCCsGAQUFBzAChj5odHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20v
# cGtpL2NlcnRzL01pY1Jvb0NlckF1dF8yMDEwLTA2LTIzLmNydDCBoAYDVR0gAQH/
# BIGVMIGSMIGPBgkrBgEEAYI3LgMwgYEwPQYIKwYBBQUHAgEWMWh0dHA6Ly93d3cu
# bWljcm9zb2Z0LmNvbS9QS0kvZG9jcy9DUFMvZGVmYXVsdC5odG0wQAYIKwYBBQUH
# AgIwNB4yIB0ATABlAGcAYQBsAF8AUABvAGwAaQBjAHkAXwBTAHQAYQB0AGUAbQBl
# AG4AdAAuIB0wDQYJKoZIhvcNAQELBQADggIBAAfmiFEN4sbgmD+BcQM9naOhIW+z
# 66bM9TG+zwXiqf76V20ZMLPCxWbJat/15/B4vceoniXj+bzta1RXCCtRgkQS+7lT
# jMz0YBKKdsxAQEGb3FwX/1z5Xhc1mCRWS3TvQhDIr79/xn/yN31aPxzymXlKkVIA
# rzgPF/UveYFl2am1a+THzvbKegBvSzBEJCI8z+0DpZaPWSm8tv0E4XCfMkon/VWv
# L/625Y4zu2JfmttXQOnxzplmkIz/amJ/3cVKC5Em4jnsGUpxY517IW3DnKOiPPp/
# fZZqkHimbdLhnPkd/DjYlPTGpQqWhqS9nhquBEKDuLWAmyI4ILUl5WTs9/S/fmNZ
# JQ96LjlXdqJxqgaKD4kWumGnEcua2A5HmoDF0M2n0O99g/DhO3EJ3110mCIIYdqw
# UB5vvfHhAN/nMQekkzr3ZUd46PioSKv33nJ+YWtvd6mBy6cJrDm77MbL2IK0cs0d
# 9LiFAR6A+xuJKlQ5slvayA1VmXqHczsI5pgt6o3gMy4SKfXAL1QnIffIrE7aKLix
# qduWsqdCosnPGUFN4Ib5KpqjEWYw07t0MkvfY3v1mYovG8chr1m1rtxEPJdQcdeh
# 0sVV42neV8HR3jDA/czmTfsNv11P6Z0eGTgvvM9YBS7vDaBQNdrvCScc1bN+NR4I
# uto229Nfj950iEkSoYIC1zCCAkACAQEwggEAoYHYpIHVMIHSMQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMS0wKwYDVQQLEyRNaWNyb3NvZnQgSXJl
# bGFuZCBPcGVyYXRpb25zIExpbWl0ZWQxJjAkBgNVBAsTHVRoYWxlcyBUU1MgRVNO
# OjNCRDQtNEI4MC02OUMzMSUwIwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBT
# ZXJ2aWNloiMKAQEwBwYFKw4DAhoDFQAoM8LvuSnUnUbrez5FdSTlglx95KCBgzCB
# gKR+MHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQH
# EwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNV
# BAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAyMDEwMA0GCSqGSIb3DQEBBQUA
# AgUA47TVtTAiGA8yMDIxMDEyMjEyNDAyMVoYDzIwMjEwMTIzMTI0MDIxWjB3MD0G
# CisGAQQBhFkKBAExLzAtMAoCBQDjtNW1AgEAMAoCAQACAg2NAgH/MAcCAQACAhGI
# MAoCBQDjtic1AgEAMDYGCisGAQQBhFkKBAIxKDAmMAwGCisGAQQBhFkKAwKgCjAI
# AgEAAgMHoSChCjAIAgEAAgMBhqAwDQYJKoZIhvcNAQEFBQADgYEAm3z4jv5zF6SK
# biTC8VTFx+DnAPbnnKSvSSQ3FgTrTy6gQVPE5u1Utzpz8TTQ5ZoFCZ4jeC97Npz0
# 4q23XFeUvIuWHfHJs2NOKY1VnxSiWiqYn6DLTb32+uTVMzp+hsjzR9a1EPRtU253
# xvK7oKnWJ0glBrGVjvtU20Qv1yI83jUxggMNMIIDCQIBATCBkzB8MQswCQYDVQQG
# EwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwG
# A1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1NaWNyb3NvZnQg
# VGltZS1TdGFtcCBQQ0EgMjAxMAITMwAAATsSG5IjUgJapQAAAAABOzANBglghkgB
# ZQMEAgEFAKCCAUowGgYJKoZIhvcNAQkDMQ0GCyqGSIb3DQEJEAEEMC8GCSqGSIb3
# DQEJBDEiBCA1IAz5DEGOg8IhDvrTuLdMmk1Nh2HE2B9pyLnoOTb5qjCB+gYLKoZI
# hvcNAQkQAi8xgeowgecwgeQwgb0EIBw25zd5xSbYAm3Zn+9CxVW3UWtY7QDGUb0W
# O2ThxEAcMIGYMIGApH4wfDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0
# b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3Jh
# dGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0IFRpbWUtU3RhbXAgUENBIDIwMTACEzMA
# AAE7EhuSI1ICWqUAAAAAATswIgQgbizQjsC+/IMrZQOVmM5grS1pm8FQh+KxQIm5
# R6UXCDAwDQYJKoZIhvcNAQELBQAEggEAgD0v8E/spBoTVIieJuYzjm27XvOt8ZPG
# 6RAg8IGJFeBTQ1vclDmsoRfJvlPAggwWw2PLZ7TPn4trCtidxg+7PrR62SniPs/B
# AH/RLcN0FTQsbI9hpW377QqCMLTAn4V7LVJzP8/ywc5DLvcJ31CpqsOivQreAx7e
# WNPkkRCN2Hg+5hp4pbAazvWrYfy0G7TmCI9joTzEhLA3O6ZRap4YaKL5pBnwvIj6
# h04ZtpiSUZ/Rqj5qXYFtqS+IHLzBcvwe/B763PLdjGm085pGVQftQr7633KwlpRt
# n7PozBD3Gm97WWIG0QOzCiqAPhRa2SVmZZ7sWAGPo8YYiqTbM99liQ==
# SIG # End signature block
